#!/usr/bin/env python3
"""Entropy and information related calculations..

**Author: Jonathan Delgado**

"""

######################## Imports ########################


import numpy as np

import stp


######################## Helper functions ########################


def _eps_filter(x):
    """ Checks if the value is within machine-epsilon of zero and maps it to 
        zero if it is the case. Useful for removing negative values in entropies that should otherwise be zero.
        
        Args:
            x (float): value to be checked.
    
        Returns:
            (float): x if the value is not within machine epsilon of zero, 0 otherwise.
    
    """
    return x if not np.isclose(x, 0, atol=9*10E-15) else 0
    

######################## Entropy calculations ########################


def entropy(p):
    """ Calculates the Shannon expression for a probability distribution.

        Args:
            p (np.ndarray): the probability distribution to calculate the entropy for

        Returns:
            (float): the entropy of p

    """
    # Since zeros do not contribute to the Shannon entropy anyways, we ignore 
        # them
    p = p[p != 0]

    KL = -np.dot(p, np.log(p))
    # Filter against machine epsilon
    return _eps_filter(KL)


def delta_entropy(R, p):
    """ Calculates the discrete time change in entropy using the entropy of p 
        evolved with R, minus the entropy of p.
    
        Args:
            R (np.ndarray): the transition matrix
            p (np.ndarray): the marginal distribution
    
        Returns:
            (float): the change in entropy
    
    """
    return entropy(step(R, p)) - entropy(p)


def relative_entropy(p, q):
    """ Calculates the Kullback-Leibler divergence, which is nonnegative and 
        vanishes if and only if the distributions coincide.
    
        Args:
            p, q (np.ndarray): the probability distributions
    
        Returns:
            (float): the relative entropy
    
    """
    if p.shape[0] != q.shape[0]:
        print('Dimensions of vectors are not equal. Cannot find relative entropy.')
        sys.exit()

    # Any values where p is zero are defined to be zero and hence do not
        # contribute to the relative entropy
    # By masking q as well we automatically skip the values that were supposed
        # to vanish with p avoiding any misalignment issues
    # Note that by masking q only where p is zero doesn't remove
        # any mismatching meaning it will still be infinite (as it should be)
        # in the case where q has a zero that p does not.
    p_filtered = p[p != 0]
    log_ratio = np.log(p_filtered / q[p != 0])

    return np.dot(p_filtered, log_ratio)


def landauer_bound(beta):
    """ Calculates the Landauer bound using beta (proportional to the inverse 
        temperature) as a parameter. This allows for freedom in setting Boltzmann's constant externally to this calculation.
        
        Args:
            beta (float): 1/kT
    
        Returns:
            (float): the Landauer bound
    
    """
    # 1/beta ln(2) = ln2 / beta = kTln2
    return np.log(2) / beta


def entropy_production(matrix, p, discrete=True):
    """ Calculates the entropy production for either discrete or continuous 
        time.
    
        Args:
            matrix (np.ndarray): the stochastic matrix, either a discrete time transition matrix or a continuous time rate matrix.

            p (np.ndarray): the marginal distribution

            discrete (bool): True if we are calculating the discrete time entropy production (nats), False if we are calculating it in continuous time (nats/time).
    
        Returns:
            (float/np.inf): the entropy production
    
    """
    log_product = matrix * np.log( matrix / matrix.T )
    # The entropy term only exists in the case of discrete time
        # it vanishes when we calculate the continuous time EP,
        # by multiplying by the boolean we include it only when
        # necessary
    EP = np.dot(log_product.sum(axis=0), p) - (entropy(p) * discrete) \
        - np.dot(stp.step(matrix, p), np.log(p))
    return EP


def entropy_flow(R, p):
    """ Calculates the discrete time entropy flow. This has not been    
        generalized to handle the continuous time entropy flow yet.
    
        Args:
            R (np.ndarray): the discrete time transition matrix
            
            p (np.ndarray): the marginal distribution
    
        Returns:
            (float): the entropy flow
    
    """
    # Vectorized calculation
    log_product = R * np.log( R / R.T )
    p_step = step(R, p)
    EF = -np.dot(log_product.sum(axis=0), p) + entropy(p_step) \
        + np.dot(p_step, np.log(p))
    return EF


######################## Entropy rates ########################


def entropy_rate(R):
    """ Calculates the asymptotic entropy rate for the provided transition 
        matrix. If the matrix is time-inhomogeneous then we return a function that generates the entropy_rate as a function of n by calculating the systems limiting distribution for each n.
        
        Args:
            R (np.ndarray/function): the transition matrix.
    
        Returns:
            (float/function): the entropy velocity.
    
    """
    if callable(R):
        return lambda n : entropy_rate(R(n))

    pst = stp.get_stationary_distribution(R, discrete=True)
    RProduct = (R * np.log(R)).sum(axis=0)

    return -np.dot(pst, RProduct)


#------------- Entry code -------------#

def main():
    print('info.py')
    

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Handles generating random objects to use in various calculations.

**Author: Jonathan Delgado**

"""

######################## Imports ########################


import sys
import numpy as np
import scipy.linalg
import bisect # for binary search

# For ProgressBar
import stp.tools.gui as gui


######################## Constructors ########################


# The random number generator to be used
# _RNG = np.random.default_rng(seed=0) # Seeded one for testing
_RNG = np.random.default_rng()
# We may want to replace zeros by machine epsilon to avoid division by zero 
# errors when calculating quantities of interest
MACHINE_EPSILON = np.finfo(float).eps


def rand_p(n=2, zeros=0):
    """ Creates a random probability distribution, currently implemented only 
        with the uniform sampling.
    
        Kwargs:
            n (int): the dimension of the desired distribution
    
            zeros (int): the number of desired zeros to be injected into the distribution. Allows for ease of testing edge cases.
    
        Returns:
            (np.ndarray): the random distribution
    
    """
    # Error checking for input amounts
    if zeros > n - 1:
        print('Too many zeros, will not allow for normalization.')
        sys.exit()
    elif zeros < 0:
        print('Invalid amount of zeros. Should be nonnegative.')
        sys.exit()
    # Creates a random distribution of length n
    p = _RNG.random(n)
    # Set the first m entries equal to 0, where m is the amount of zeros 
        # requested
    if zeros != 0:
        p[:zeros] = 0
        # Shuffle the contents to avoid just having the same first entries as 0
        np.random.shuffle(p)
    # Return the distribution normalized by its 1-norm    
    return p / np.sum(p)


def rand_rate_matrix(n=2):
    """ Generates a random, time-independent, n x n rate matrix consisting of 
        probabilities per unit time. Column normalized.
    
        Kwargs:
            n (int): the number of states the matrix will correspond to. Relates to the dimensions of the matrix.
    
        Returns:
            (np.ndarray): the rate matrix.
    
    """
    # Generates a random n x n matrix
    W = _RNG.random( (n, n) )

    np.fill_diagonal(W, 0)
    # Normalize the columns to sum to zero.
    for y in range(n):
        W[y, y] = -np.sum(W[:,y])

    return W


def rate_to_transition_matrix(W, time_step):
    """ Converts a rate matrix to a transition matrix assuming a constant 
        control parameter during the duration of time_step.
        
        Args:
            W (np.ndarray): the rate matrix

            time_step (float): the time step
    
        Returns:
            (np.ndarray): the transition matrix
    
    """
    return scipy.linalg.expm(W * time_step)


def rand_transition_matrix(n=2, time_step=1.0):
    """ Generates a random, time-independent, discrete-time, transition matrix
        by first generating a random rate matrix and then matrix exponentiating it to incorporate the time step as an additional parameter.
    
        Kwargs:
            n (int): the number of states the matrix will correspond to. Relates to the dimensions of the matrix.

            time_step (float): the time step: the interval of time between observations.
    
        Returns:
            (np.ndarray): the n x n transition matrix
    
    """
    return rate_to_transition_matrix(rand_rate_matrix(n), time_step)


def self_assembly_rate_matrix(energy=1, c=1, M=1, T=1):
    """ Generates a self-assembly rate matrix for a 3-state system following:
        https://aip.scitation.org/doi/10.1063/1.3662140.
        
        Kwargs:
            energy (float): the negative of the "optimally bound level of energy"

            c (float): "concentration-like variable"

            M (int): the degeneracy of the misbound level

            T (float/function): the temperature as a float or a function which returns the temperature as a function of time. In the latter case the returned rate matrix will be a function which provides a np.ndarray as a function of time.
    
        Returns:
            (np.ndarray/function): the time-independent rate matrix as a numpy array in the case where the temperature is constant. Otherwise returns a function corresponding to the time-dependent rate matrix.
    
    """
    if not isinstance(M, int):
        print('The degeneracy of the misbound level, M, must be an integer.')
        sys.exit()

    if not callable(T) and T <= 0:
        print('Must provide a nonnegative temperature.')
        sys.exit()

    if callable(T):
        # T is a time-dependent temperature
        # Provide the time-dependent rate matrix
        return lambda t : self_assembly_rate_matrix(energy=energy, c=c, M=M, T=T(t))

    ### Main body ###

    alpha = np.exp(-energy / (2 * T))
    W = np.array([
        [   -c * (M + 1),   alpha,      alpha**2    ],
        [   c * M,          -alpha,     0           ],
        [   c,              0,          -alpha**2   ]
    ])

    return W


def self_assembly_transition_matrix(energy=1, c=1, M=1, T=1, time_step=1):
    """ Generates a self-assembly, discrete time, transition matrix for a 
        3-state system following: https://aip.scitation.org/doi/10.1063/1.3662140. Done assuming any external control parameter is fixed for the duration of the time_step. This matrix is step dependent, so conversions will be done to time to calculate the current temperature.
    
        Kwargs:
            energy (float): the negative of the "optimally bound level of energy"

            c (float): "concentration-like variable"

            M (int): the degeneracy of the misbound level

            T (float/function): the temperature as a float or a function which returns the temperature as a function of time. In the latter case the returned transition matrix will be a function which provides a np.ndarray as a function of time. This external control parameter will be fixed for the duration of the time_step.

            time_step (float): the time step
    
    
        Returns:
            (np.ndarray/function): the transition matrix
    
    """
    if callable(T):
        # Convert n to time.
        # This also serves to fix the temperature between observations to change
            # the control parameter discretely.
        old_T = T
        T = lambda n : old_T(n * time_step)

    return rate_to_transition_matrix(
            self_assembly_rate_matrix(energy=energy, c=c, M=M, T=T),
            time_step
        )


######################## Probability Operations ########################


def step(matrix, p):
    """ Evolves a probability distribution one step forward by computing the 
        matrix multiplication between matrix and p. In the case of the matrix being a rate matrix the output is the time-derivative of p.
        
        Args:
            matrix (np.ndarray): the transition or rate matrix

            p (np.ndarray): the marginal distribution
    
        Returns:
            (np.ndarray): the evolved marginal
    
    """
    return np.matmul(matrix, p)
    

def get_stationary_distribution(matrix, discrete=True):
    """ Calculates the stationary distribution of a transition or rate matrix. 
        Credit: https://stackoverflow.com/questions/31791728/python-code-explanation-for-stationary-distribution-of-a-markov-chain.
        
        Args:
            matrix (np.ndarray/function): the transition or rate matrix
    
        Kwargs:
            discrete (bool): True if the provided matrix is a discrete time transition matrix, False if the provided matrix is a continuous time rate matrix
    
        Returns:
            (np.ndarray/function): the limiting distribution (as a function of time in the case where the matrix is too)
    
    """
    if callable(matrix):
        # Return the limiting distribution as a function of time
        return lambda t : get_stationary_distribution(matrix(t), discrete=discrete)

    if not discrete:
        # Use the simplified method for the rate matrix, not as robust
            # Make sure to check output
        _, eigenvec = np.linalg.eig(matrix)
        print('Returning experimental right eigenvector to the rate matrix. Always check the vector is correct!')
        return np.abs(eigenvec[:, 0])

    S, U = np.linalg.eig(matrix)
    # Added the need to take the real part to avoid +0j added to each term
    # and raising a complex error
    # Added absolute value since I was occasionally getting the negative p
    stationary = np.absolute(
        np.array(U[:, np.where(np.abs(S - 1.) < 1e-8)[0][0]].flat).real
    )
    return stationary / np.sum(stationary)


def get_path_probability(R, p, path):
    """ Calculates the probability of observing a provided path using the a
        (potentially time-inhomogeneous) transition matrix, the initial marginal distribution, and the path itself.
        
        Args:
            R (function/np.ndarray): the transition matrix. A function of the observation step if the matrix is time-inhomogeneous. That is, R = R(n). In which case Pr( x <- y ) = R(1)[x,y] * p[y]. Provide a numpy matrix in the case of a time-homogeneous transition matrix.

            p (np.ndarray): the marginal distribution.

            path (list/np.ndarray): the path.
    
        Returns:
            (float): the probability of observing this path.
    
    """
    # Check if R is written as a numpy matrix, if so, convert it to a function
        # of the observation step that simply returns the constant matrix.
    if isinstance(R, np.ndarray):
        matrix = R
        R = lambda n: matrix

    # Comprehension to collect transition probabilities for vectored product
    total_jump_prob = np.array([
        R(i + 1)[path[i+1], path[i]]
        for i in range( len(path) - 1 )
    ]).prod()

    return total_jump_prob * p[path[0]]


######################## Path space sampling ########################


def complete_path_space(n, path_length):
    """ Generates the entire path space as a matrix with each row corresponding 
        to a path and each column corresponding to an observation step.
        
        Args:
            n (int): the number of states in the state space

            path_length (int): the length of each path
    
        Returns:
            (np.ndarray): the entire path space. Will be a matrix of size: n^path_length x path_length.
    
    """
    paths = []

    num_paths = n**path_length

    bar = gui.ProgressBar(num_paths, width=300, title='Generating path space...')
    for i in range(num_paths):
        path = np.base_repr(i, base=n)
        path = ( path_length - len(path) )*'0' + path
        paths.append(
            # Convert the number from decimal to base n.
            # This will be in a string format, turn each digit into a separate
                # int in a list, this will be a path.
                [ int(state) for state in path ]
            )

        bar.update()

    bar.finish()

    return np.array(paths)


def KMC(W, p, num_paths, final_time, time_step=1, discrete=True, degenerate_threshold=0.985):
    """ A Rejection-free Kinetic Monte Carlo (KMC) algorithm for simulating 
        the time evolution of a system, where some processes can occur with known rates W = W(t). From: https://en.wikipedia.org/wiki/Kinetic_Monte_Carlo.
        
        Args:
            W (np.ndarray/function): rate matrix for np.ndarray. If function is provided then W is the function of time that provides a rate matrix (W = W(t)) for each moment in time. Will be used to implement driven systems. If W is just a rate matrix then W is just the time-homogeneous rate matrix.

            p (np.ndarray): the initial marginal distribution.

            num_paths (int): the number of paths to sample.

            final_time (float): the final time.
    
        Kwargs:
            time_step (float): the interval between changing of rate matrix (changing of control parameter), and the delay between observations.

            discrete (bool): True if the the simulation is for discrete time and hence the control parameter will be switched accordingly. If False the algorithm will be done for continuous time and will return the time series too.

            degenerate_threshold (float): the fraction of paths that we require to be unique. If not satisfied KMC will be run again with an increased threshold.
    
        Returns:
            (np.ndarray): matrix of paths sampled via KMC.
    
    """
    #------------- Helper functions -------------#


    def _KMC_jump(cumulatives, state):
        """ Helper function which takes cumulatives, a state, and generates 
            the next likely jump via the KMC algorithm and the time it takes to observe the jump. Useful for discretizing the KMC.
            
            Args:
                cumulatives (np.ndarray): the KMC's cumulatives.

                state (int): the current state.
        
            Returns:
                (2-tuple): (next_state, delta time)
        
        """
        # https://en.wikipedia.org/wiki/Kinetic_Monte_Carlo.
        # The cumulative relevant to the current state
        relevant_cumulative = cumulatives[:, state]
        u = _RNG.uniform(0, 1)
        scaled_total_rate = u * cumulatives[-1, state]
        next_state = bisect.bisect_left(relevant_cumulative, scaled_total_rate)
        u = _RNG.uniform(0, 1)
        jump_time = -np.log(u) / relevant_cumulative[-1]

        return next_state, jump_time


    def _KMC_discrete(W, p, num_paths, final_time, time_step=1):

        #------------- Initialization -------------#

        # Read off the size of the state space.
        n = W(0).shape[0]
        state_options = np.arange(n)
        # Read off the path length for the full time series observations given 
            # the final time and the delay between observations.
        # For final_time, time_step, and path_length, the relationship is:
        # time_step * (path_length - 1) = final_time, which implies that
        path_length = int(final_time / time_step + 1)

        # Initialize the matrix storing the paths
        paths = np.full( (num_paths, path_length), -1, dtype='int64' )
        # Set the first column of the matrix to be initial state weighted by p.
        paths[:, 0] = _RNG.choice(state_options, p=p, size=num_paths)

        # Fill each column
        for k in range(1, path_length):
            # This will find the current time of the control parameter to fix
                # the rate matrix since this is discrete time.
            current_time = k * time_step
            # Fix the rate matrix
            current_W = W(current_time)

            #------------- Main algorithm -------------#

            # We need to remove the normalization for the algorithm and just 
                # look at the positive rates.
            posW = current_W.copy()
            np.fill_diagonal(posW, 0)
            # Matrix of partial sums, last one being the total rate
            # Each row corresponds to the same final index for the partial sum
            # Each column corresponds to which rates were summed together
            cumulatives = np.array( [ [ 
                np.sum(posW[:i+1, x])
                for i in range(n)
            ]   for x in range(n)
            ] ).T

            # Run through the path matrix.
            for path in paths:
                current_state = path[k - 1]
                # The state we are currently in is the last nonnegative 
                    # element of our path list (since the path matrix is
                    # initialized with -1)
                # current_state = path[path >= 0][-1]
                
                total_jump_times = 0
                # Prepare for the while loop
                next_state = current_state
                # Keep evolving the state until it fits what would be observed
                    # in a discrete time process.
                while total_jump_times < time_step:
                    potential_jump, jump_time = _KMC_jump(cumulatives, next_state)
                    
                    # Note that if the next jump takes longer than our 
                        # time_step we won't observe it. Since we will
                        # make the measurement before this happens.
                    if total_jump_times + jump_time >= time_step:
                        break

                    total_jump_times += jump_time
                    next_state = potential_jump

                path[k] = next_state

        ### Degenerate paths check ###
        # Check for degenerate paths and sample again if needed.
        paths = np.unique(paths, axis=0)
        if paths.shape[0] / num_paths < degenerate_threshold:
            # Generate a rate matrix of perturbations
            deltaW = rand_rate_matrix(n) / 100
            Wprime = lambda t : W(t) + deltaW

            # Run KMC again with an increased degeneracy threshold and 
                # a perturbed rate matrix.
            new_paths = KMC(Wprime, p, num_paths, final_time, time_step=time_step, discrete=discrete, degenerate_threshold=degenerate_threshold-10E-9)

            # Combine the path data
            paths = np.unique( np.vstack((paths, new_paths)), axis=0 )
            
        # Return only the requested number of paths (in case we have extras
            # from previous degeneracies)
        return paths[:num_paths]

    #------------- Body -------------#

    if isinstance(W, np.ndarray):
        # Hold the rate matrix in memory.
        constW = W
        # Let the time-dependent rate matrix just be the constant matrix.
        W = lambda t: constW

    if discrete:
        return _KMC_discrete(W, p, num_paths, final_time, time_step=time_step)

    print('Continuous time KMC has not been implemented yet.')
    print('Exiting...')
    sys.exit()


def direct_sampling(R, p):
    """ Samples the path space directly to reflect the dynamics given by the 
        initial marginal and the transition matrix.
        
        Args:
            arg1 (arg1 type): arg1 description.
    
        Kwargs:
            karg1 (arg1 type): arg1 description.
    
    
        Returns:
            (None): none
    
    """
    print('Direct sampling has not been implemented yet.')
    

######################## Entry Code ########################


def main():
    print('stochastic.py')
    
    
if __name__ == '__main__':
    main()


#!/usr/bin/env python3
"""Calculates a typical set for provided systems, incorporating various methods of path space sampling to construct a representative set in a reasonable amount of time.

**Author: Jonathan Delgado**

"""

######################## Imports ########################


import numpy as np

import stp
import stp.info
# For ProgressBar
import stp.tools.gui as gui


######################## Helper functions ########################


def probability_captured(typical_set, sampled_set):
    """ Calculates the amount of probability captured by the typical set
        and performs any relevant analysis.
        
        Args:
            typical_set (list): the typical set.

            sampled_set (list): the sampled path space.
    
        Returns:
            (float): the probability captured by the typical set.
    
    """
    TS_probability = sum( [ entry[1][-1] for entry in typical_set ] )
    print( f'Probability captured by the sampled typical set: {TS_probability}.' )
    
    sampled_probability = sum( [ entry[1][-1] for entry in sampled_set ] )
    print( f'Total probability sampled: {sampled_probability}.' )
    
    print( f'Percentage of sampled probability that is typical: {TS_probability / sampled_probability * 100:.1f}%.' )

    return TS_probability


def verify_TS(typical_set, R, epsilon):
    """ Tests Theorem 3.1.2 of Cover & Thomas.
        
        Args:
            typical_set (list): a list of 2-tuples where the first component is the path, and the second component is the path probability as a function of n.All paths in this list are typical.

            R (np.ndarray): the transition matrix used to calculate the limiting distribution for the system and hence its entropy rate. If it is time-inhomogeneous, the typical set will be used to pull out the relevant transition matrix.

            epsilon (float): the epsilon being used
    
        Returns:
            (None): none
    
    """
    num_typical = len(typical_set)
    # All path lengths are the same so just check how long the first path is.
    if num_typical > 0:
        n = len(typical_set[0][0])
    else:
        # n can just be assigned since the relevant theorem will be False.
        n = 1000

    if callable(R):
        # The matrix is time-inhomogeneous.
        # We just need the transition matrix corresponding to the final entropy
            # rate
        R = R(n)

    # Theorem 3.1.2:
    # 2. Pr A(n) > 1 − epsilon for n sufficiently large.
    # 3. |A(n)| ≤ 2^{n( H(X)+epsilon )}
    # 4. |A(n)| ≥ (1 − epsilon)2^{n( H(X)−epsilon )} for n sufficiently large.

    # Get the probability captured by the typical set at the final observation
    TS_probability = sum( [ entry[1][-1] for entry in typical_set ] )

    entropy_rate = info.entropy_rate(R)

    ### Printing of test results. ###

    print('NOTE: Only Theorem 3.1.2.3 is REQUIRED to be true for these simulations. It is not necessarily incorrect for 3.1.2.2 or 3.1.2.4 to be False.')
    print()
    print(f'Probability of observing a typical path: {TS_probability}.')
    print(f'1 - epsilon: {1 - epsilon}.')
    print(f'Theorem 3.1.2.2 satisfied: {TS_probability > 1 - epsilon}.')
    print()
    print(f'Number of typical paths: {num_typical}.')
    print(f'Entropy rate: {entropy_rate}.')
    upper_bound = 2**(n * (entropy_rate + epsilon))
    print(f'Theorem 3.1.2.3 satisfied: {upper_bound > num_typical}.')
    print()
    lower_bound = (1-epsilon) * 2**(n * (entropy_rate - epsilon))
    print(f'Theorem 3.1.2.4 satisfied: {num_typical > lower_bound}.')


######################## Main body ########################


def typical_set(R, p, path_space, time_step=1, epsilon=1):
    """ Samples a path space and generates a typical set on this sampled path 
        space.

        Args:
            R (np.ndarray/function): the transition matrix, time-dependent if provided as a function

            p (np.ndarray): the initial marginal distribution

            path_space (np.ndarray): the portion of the path space to use.
    
        Kwargs:
            time_step (float): the time between observations

            epsilon (float): the epsilon neighborhood to consider paths to be typical within
    
        Returns:
            (2-tuple): the first component being the typical set as a list of 2-tuples, each tuple being the (path, path probability vs n), the second component of the returned tuple is the total path space as a list of 3-tuples, each tuple being (path, path probability vs n, True if typical False if a typical as a function of n). In this case a path is added to the typical set if its final path entropy rate is within the typical set bounds. While a path may be atypical by the end of the observation, it could have been typical before, and this information is encoded in the last component as a list of bools.
    
    """
     
    #------------- Data preparation -------------#

    # Convert the transition matrix to add time-dependence as a constant matrix
        # if a np.ndarray was provided
    if isinstance(R, np.ndarray):
        oldR = R
        R = lambda n : oldR

    # The number of states
    n = len(p)

    final_time = time_step * (path_length - 1)
    num_paths, path_length = path_space.shape

    # Save the initial distribution
    pinit = p

    # List of 2-tuples representing the typical set
        # (path_probability, forward path)
    typical_set = []
    # List of 3-tuples representing the sampled path space
        # (path_probability, forward path, is_typical)
    sampled_set = []

    path_probabilities = np.zeros(paths.shape)

    # Used for the bounds
    entropy_rate = info.entropy_rate(R)
    entropy_rate_vs_time = [
        entropy_rate(i)
        for i in range(path_length)
    ]

    #------------- Data gathering -------------#

    bar = gui.ProgressBar(path_length * num_paths, width=300, title='Gathering data...')
    
    ### Quantities versus time ###
    for current_path_length in range(2, path_length + 1):
        # The data index
        i = current_path_length - 1
        # Since the marginals are zero-indexed as are the paths
        step_index = current_path_length - 2

        currentR = R(current_path_length - 1)
        # Propagate the marginal one step and save it separately
            # for quantities like the temporal coarse graining term
        pstep = step(currentR, p)

        ### Path probability calculations ###
        for x, path in enumerate(paths):

            current_state = path[step_index]
            jump_state = path[step_index + 1]

            # Forward calculations
            # Recursive calculation to save time
            last_joint = path_probabilities[x, i - 1]
            jump_prob = currentR[jump_state, current_state]
            path_probabilities[x, i] = last_joint * jump_prob

            bar.update()


        # Finished data gathering for this iteration, propagate marginal
            # forward in time
        p = pstep


    bar.finish()


    #------------- Typical set bounds and generation -------------#

    TS_upper = entropy_rate_vs_time + epsilon
    TS_lower = entropy_rate_vs_time - epsilon

    path_entropy_rates = -np.log(path_probabilities) / ns

    # Identify the paths that are typical and atypical
    # Populate the sets
    for path_index, path_entropy_rate in enumerate(path_entropy_rates):

        # Information for typical set and sampled_set
        path_info = [
            paths[path_index],
            # The list of the probability this path observed with each
                # added state
            path_probabilities[path_index],
            # List of bools where the True means the path is typical at this
                # step, False if it is atypical
            # If the last element is True then this tuple is added to the
                # typical set
            (TS_lower < path_entropy_rate) & (path_entropy_rate < TS_upper)
        ]

        # The path is considered typical at the end of the observation
        # Append the 3-tuple to the typical set
        if path_info[-1][-1]:
            typical_set.append( tuple(path_info[:-1]) )

        sampled_set.append( tuple(path_info) )


    #------------- Analysis & Printing -------------#

    percent_typical = f'{len(typical_set) / num_paths * 100:.1f}'
    print(f'{percent_typical}% of tested paths are typical.')


    return typical_set, sampled_set


#------------- Entry code -------------#


def main():
    print('typical_set.py')
    

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Top-level init file to make package available for import.

**Author: Jonathan Delgado**

Handles importing and setting project-wide fields.

"""

__author__ = 'Jonathan Delgado'
__version__ = '0.0.1.0'

# Make all functions from stochastic accessible on import of stp
from .stochastic import *


#!/usr/bin/env python3
"""GUI element handler.

**Author: Jonathan Delgado**

Handles creating GUI elements such a graphical loading bars for long processes.

"""
from tkinter import ttk
import tkinter as tk
import time

######################## GUI ########################

def _center_window(master, width, height):
        screen_width = master.winfo_screenwidth()
        screen_height = master.winfo_screenheight()
        # Calculate (x, y) coordinates for the Tk window
        x = (screen_width/2) - (width/2)
        y = (screen_height/2) - (height/2)
        # Set the dimensions of the screen and where it is placed
        master.geometry('%dx%d+%d+%d' % (width, height, x, y))


class ProgressBar:
    """ Tkinter progress bar.
    
        Object which will handle creating a Tkinter progress window for the purpose of showing as a graphical loading bar.
        
        Attributes:
            MAX_VALUE: the max value of steps before completing the bar.

            width: the width of the GUI bar.

            height: the height of the GUI bar.
            
            title: the window title.
    """

    def __init__(self, MAX_VALUE, width=200, height=40, title='Loading...'):
        """ Initialize object to hold important quantities and create/style the bar. """


        # Must know maximum to calculate percentages
        self._MAX_VALUE = MAX_VALUE
        # Initialize value property
        self.value = 0
        # self.master = tk.Toplevel()
        self.master = tk.Tk()
        self.bar = ttk.Progressbar(self.master, orient=tk.HORIZONTAL, length=100, mode='determinate')
        self.width = width; self.height = height
        self.title = title

        self._style()
        _center_window(self.master, self.width, self.height)


    def _style(self):
        """ Handles designing the bar. """

        self.bar.pack(fill='both', expand=True)
        self.set_title(self.title)
        self.master.style = ttk.Style()
        self.master.style.theme_use('classic')


    def set_title(self, title):
        """ Updates the bar's title.
            
            Args:
                title (str): the new title.

            Returns:
                (None): none
        
        """
        self.title = title
        self.master.wm_title(self.title)


    def update(self, amount=1):
        """ Main function for interacting with the bar. Handles updating progress.
            
            Kwargs:
                amount (int): the amount to update the progress by.
        
        """
        self.value += amount

        # print( f'Progress update counter: \n{self.value}' )
        status = self.value / self._MAX_VALUE * 100
        self.bar['value'] = status
        # time.sleep(1) # Uncomment to see the progress bar update slowly

        # Updates the window itself
        self.master.update()


    def next(self):
        """ Mask for update to use same syntax in the case of using instead of ShadyBar for example. """
        self.update()


    def finish(self):
        """ Finishes the bar. Briefly shows it has been completed before safely destroying the Tkinter window. """
        self.bar['value'] = 100
        self.master.update()
        # Briefly show a completed bar
        time.sleep(0.2)
        # Close the window
        self.master.destroy()




#------------- Entry code -------------#

def main():
    print('gui.py')
    

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
"""Setup for pip install stp.

**Author: Jonathan Delgado**

"""
from setuptools import setup, find_packages

setup(
    name='stp',
    version='0.0.1.0',
    description='Stochastic Thermodynamics in Python',
    license='MIT',
    url='https://github.com/otanan/STP',
    author='Jonathan Delgado',
    author_email='jonathan.delgado@uci.edu',
    keywords=['stochastic', 'thermodynamics', 'python', 'information'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],
    download_url='https://github.com/otanan/STP/archive/refs/tags/0.0.1.0.tar.gz',
    # packages=find_packages(),
    packages=[
        'stp',
        'stp.tools',
    ],
    install_requires=[
        # External packages
        'numpy',
    ],
)

